#!/usr/bin/env ruby

# bootstrap

require 'rice/irc'
require 'rice/observer'

require 'net/http'

require 'nkf'
require 'kconv'

require 'const'

include RICE::Command

STDOUT.sync = true
STDERR.sync = true

print "booting mirc...\n"

o = RICE::SimpleClient.new(NICK, USER, REAL, PASS,
                           CHANNEL.collect {|i| NKF::nkf('-Ej', i)})
c = RICE::Connection.new(SERV, PORT)

c.add_observer(o)

class IRCContext
  attr_reader :subject, :message, :nick, :user, :time
  attr_accessor :target

  def initialize(subject, message)
    @subject = subject
    @target = message.params[0]
    @message = message.params[1..-1].join(' ')
    message.prefix =~ /^([^!]*)!(.*)/
    @nick = $1
    @user = $2
    if (@target == NICK)
      @target = @nick
    end
    @time = Time.now
  end

  def notice(msg)
    @subject.push(RICE::Command::notice(@target, msg))
  end
end

class Mirc
  @@mircs = Hash.new

  def Mirc.get(subject, message)
    target = message.params[0]
    if (!@@mircs.key?(target))
      @@mircs[target] = Mirc.new(subject, message)
    end
    @@mircs[target]
  end

  def initialize(subject, message)
    @context = IRCContext.new(subject, message)

    @conf = Hash.new
    @conf['use_color'] = '1'
    @conf['color'] = '13'

    # ad-hoc!
    @plugins = Hash.new
    load_plugin('serialize')

    load_plugins
  end

  def load_plugin(name)
    open("plugins/#{name}.rb") do |src|
      self.instance_eval(src.read)
    end

    if (self.methods.include?("#{name}_init"))
      self.method("#{name}_init").call
    end

    @plugins[name] = [
      if (name =~ /!/)
        self.method("#{name.sub(/!/, '')}_desc!").call
      else
        self.method("#{name}_desc").call
      end,
      self.method(name),
      Time.now
    ]
  end

  def load_plugins
    print "load plugins...\n"

    @plugins = Hash.new
    Dir.open('plugins') do |dir|
      dir.each do |d|
        begin
          if (d.to_s !~ /^([^\.]*)\.rb$/)
            next
          end

          load_plugin($1.downcase)
        rescue
          p $!.to_s
          print "#{$!.backtrace.join("\n")}\n"
        end
      end
    end

    #print "load plugins done.\n"
  end

  def invoke_impl(cmd, arg)
    if (arg =~ /^([a-zA-Z!]+)\>(.*)/)
      r = invoke_impl($1, $2)
      if (r.class == String)
        arg = r
      end
    end

    rbname = "plugins/#{cmd}.rb"
    if (File.exist?(rbname) &&
          (!@plugins.key?(cmd) || File.mtime(rbname) > @plugins[cmd][2]))
      print "#{rbname} is modified. reload it.\n"
      load_plugin(cmd)
    end

    if (@plugins.key?(cmd))
      res = @plugins[cmd][1].call(arg)
      return res
    else
      return nil
    end
  end

  def invoke(context, cmd, arg)
    begin
      @context = context

      if (cmd =~ /^([^#]+)(#.*)$/)
        cmd = $1
        @context.target = $2
      end

      res = invoke_impl(cmd, arg)

      if (res)
        if (!res || res == "")
        elsif (res.class == Array)
          res.each do |r|
            notice(r)
          end
        else
          notice(res)
        end
        print "#{cmd}(#{arg}) => #{res}\n"
        return true
      else
        print "#{cmd}: command not found\n"
        #      subject.push(privmsg(message.prefix, "#{cmd}: command not found"))
        return false
      end
    rescue
      p $!.to_s
      print "#{$!.backtrace.join("\n")}\n"
      notice("(ء) 㳰Фޤ")
      notice($!)
      return true
    end
  end

  def notice(msg)
    msg = NKF::nkf('-Ej', msg)
    if (@conf['use_color'] == '1')
      msg = 3.chr + "#{@conf['color']} " + msg
    end
    @context.notice(msg)
  end

  def toeuc(str)
    Kconv::toeuc(str)
  end
  def j2e(str)
    NKF::nkf('-Je', str)
  end
  def e2j(str)
    NKF::nkf('-Ej', str)
  end

  def http_get(url)
    url =~ /http:\/\/([^\/]+)(\/.*)$/
    serv = $1
    path = $2
    Net::HTTP.version_1_2
    if (PROXY_HOST)
      Net::HTTP::Proxy(PROXY_HOST, PROXY_PORT)
    else
      Net::HTTP
    end.start(serv) do |http|
      resp, data = http.get(path, nil)
      while (resp.code =~ /^5/)
        sleep(3)
        resp, data = http.get(path, nil)
      end
      return toeuc(data)
    end
  end
end

CTCP_VERSION = 1.chr + "VERSION" + 1.chr
CTCP_CLIENTINFO = 1.chr + "CLIENTINFO" + 1.chr

def o.response_for_join(subject, message)
  Mirc.get(subject, message)
end

def o.response_for_privmsg(subject, message)
  begin
    context = IRCContext.new(subject, message)

    msg = context.message

    if (msg == CTCP_VERSION && context.target[0] != '#')
      print "CTCP VERSION from #{context.nick}\n"

      name = 'mircbot'
      ver = '0.0.3'
      env = if (HAVE_UNAME)
              `uname -s -r`.chomp.sub(/ /, '/')
            else
              RUBY_PLATFORM
            end + " Ruby/" + RUBY_VERSION

      context.notice(1.chr + "VERSION #{name} #{ver} (#{env})" + 1.chr)

      return
    elsif (msg == CTCP_CLIENTINFO && context.target[0] != '#')
      print "CTCP CLIENTINFO from #{context.nick}\n"

      context.notice(1.chr + "CLIENTINFO :CLIENTINFO VERSION" + 1.chr)

      return
    end

    cmd = nil
    arg = nil
    forward_cmd = false
    if (msg =~ /^([a-zA-Z!]+(#[^#]+)?)\>(.*)/)
      cmd = $1
      arg = $3
      forward_cmd = true
    elsif (msg =~ /^(.*?)>([a-zA-Z!]+)$/)
      cmd = $2
      arg = $1
    else
      return
    end
    ret = Mirc.get(subject, message).invoke(context, cmd, arg)

    if (!ret && forward_cmd)
      if (msg =~ /^(.*?)>([a-zA-Z!]+)$/)
        cmd = $2
        arg = $1
        Mirc.get(subject, message).invoke(context, cmd, arg)
      end
    end
  rescue
    p $!.to_s
    print "#{$!.backtrace.join("\n")}\n"
  end
end

def o.response_for_invite(subject, message)
  nick = message.prefix
  chan = message.params[1]

  print "invited by #{nick} to #{chan}\n"

  subject.push(join(chan))
end

begin
  c.start
rescue RICE::Connection::Closed
  p $!.to_s
  print "#{$!.backtrace.join("\n")}\n"
end
